/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RemoteLSSSimulation.h"

#include <InstrumentManager/InstrumentManager.h>
#include <boost/lexical_cast.hpp>
#include <fstream>
#include <iostream>
#include "controllers/common/family/Families.h"
#include <InstrumentManager/InstrumentManager.h>
#include "controllers/common/acquisition/ExperimentData.h"
#include "controllers/common/utilities/DataParams.h"
#include "controllers/common/axis/AxisController.h"
#include "controllers/common/state/stateio/AttenuatorController.h"
#include "controllers/common/state/stateaxis/StateAxisController.h"
#include "controllers/lss/common/LSSWavelength.h"
#include "controllers/lss/common/CollimationProperties.h"
#include "controllers/common/acquisition/Count.h"
#include "controllers/common/acquisition/detector/DetectorElement.h"
#include <boost/tokenizer.hpp>

namespace remotelsssimulation {

using namespace std;
using namespace cameo;
using namespace boost;

const std::string RemoteLSSSimulation::TYPE = "remote_lss_simulation";

const std::string RemoteLSSSimulation::MATLAB_APPLICATION = "lss-simulation-matlab";
const std::string RemoteLSSSimulation::RESPONDER_NAME = "lss_simulation_controller";

const std::string RemoteLSSSimulation::SAMPLE_TYPE = "sample";
const std::string RemoteLSSSimulation::BACKGROUND_TYPE = "background";
const std::string RemoteLSSSimulation::BLOCKEDBEAM_TYPE = "blocked";

const std::string RemoteLSSSimulation::D11_COLLIMATION_TYPE = "d11collimationsetting";
const std::string RemoteLSSSimulation::D22_COLLIMATION_TYPE = "d22collimationsetup";
const std::string RemoteLSSSimulation::D33_COLLIMATION_TYPE = "lsscollimationsetting";

/*
 * constructor
 */
RemoteLSSSimulation::RemoteLSSSimulation(const std::string& name) :
		ExperimentController(name), controller::Start(this), controller::Init(this) {

	setFamily(family::ACQUISITION, family::SETTING);

	serverEndpoint.init(this, SAVE, "cameo_server");
	initialized.init(this, NOSAVE, "initialized");
	applicationState.init(this, NOSAVE, "matlab_state");

	useSimulatedData.init(this, SAVE, "use_simulated_data");

	model_name.init(this, NOSAVE, "model_name");
	model_type.init(this, NOSAVE, "model_type");
	model_number.init(this, NOSAVE, "model_number");
	pnames.init(this, NOSAVE, "pnames");
	structname.init(this, NOSAVE, "structname");
	fn_eval.init(this, NOSAVE, "fn_eval");
	parameters.init(this, NOSAVE, "parameters");

	nbSampleModels.init(this, NOSAVE, "nb_sample_models");
	sampleModelName.init(this, NOSAVE, "sample_model_name");
	displaySampleModelName.init(this, NOSAVE, "display_sample_model_name");
	displaySampleModelType.init(this, NOSAVE, "display_sample_model_type");
	displaySampleNbParameters.init(this, NOSAVE, "display_sample_nb_parameters");
	displaySampleParametersNames.init(this, NOSAVE, "display_sample_parameters_names");
	displaySampleParametersValues.init(this, NOSAVE, "display_sample_parameters_values");

	// Fix size
//	displaySampleParametersNames.resize(8);
//	displaySampleParametersValues.resize(8);

	nbBackgroundModels.init(this, NOSAVE, "nb_background_models");
	backgroundModelName.init(this, NOSAVE, "background_model_name");
	displayBackgroundModelName.init(this, NOSAVE, "display_background_model_name");
	displayBackgroundModelType.init(this, NOSAVE, "display_background_model_type");
	displayBackgroundNbParameters.init(this, NOSAVE, "display_background_nb_parameters");
	displayBackgroundParametersNames.init(this, NOSAVE, "display_background_parameters_names");
	displayBackgroundParametersValues.init(this, NOSAVE, "display_background_parameters_values");

	// Fix size
//	displayBackgroundParametersNames.resize(8);
//	displayBackgroundParametersValues.resize(8);

	nbBlockedBeamModels.init(this, NOSAVE, "nb_blockedbeam_models");
	blockedBeamModelName.init(this, NOSAVE, "blockedbeam_model_name");
	displayBlockedBeamModelName.init(this, NOSAVE, "display_blockedbeam_model_name");
	displayBlockedBeamModelType.init(this, NOSAVE, "display_blockedbeam_model_type");
	displayBlockedBeamNbParameters.init(this, NOSAVE, "display_blockedbeam_nb_parameters");
	displayBlockedBeamParametersNames.init(this, NOSAVE, "display_blockedbeam_parameters_names");
	displayBlockedBeamParametersValues.init(this, NOSAVE, "display_blockedbeam_parameters_values");

	// Fix size
//	displayBlockedBeamParametersNames.resize(8);
//	displayBlockedBeamParametersValues.resize(8);

	m_driver.init(this, "driver");

	initialized = false;
	m_displaySampleModel = 0;
	m_displayBackgroundModel = 0;
	m_displayBlockedBeamModel = 0;
	m_CountSpy = nullptr;
}

/*
 * constructor
 */
RemoteLSSSimulation::RemoteLSSSimulation(const RemoteLSSSimulation& controller) :
		ExperimentController(controller), controller::Start(this), controller::Init(this) {
	// Empty
	initialized = false;
	m_displaySampleModel = 0;
	m_displayBackgroundModel = 0;
	m_displayBlockedBeamModel = 0;
	m_CountSpy = nullptr;
}

/*
 * destructor
 */
RemoteLSSSimulation::~RemoteLSSSimulation() {
	// Empty
}

/*
 * postConfiguration
 */
void RemoteLSSSimulation::postConfiguration() {

	if (m_server.get() == 0) {
		// Create a new server if it is not already created.
		// In case of a simulated server, avoid the real remote endpoint.
		if (serverEndpoint() == "" || m_driver->state() != REAL_DEVICE_CONTAINER) {
			m_server.reset(new Server(application::This::getServer().getEndpoint()));
		} else {
			m_server.reset(new Server(serverEndpoint()));
		}
	}
	registerRefresher(displaySampleModelName, &RemoteLSSSimulation::refreshDisplaySampleModelProperty, this);
	registerRefresher(displaySampleParametersValues, &RemoteLSSSimulation::refreshDisplaySampleParametersValuesProperty, this);
	registerRefresher(displayBackgroundModelName, &RemoteLSSSimulation::refreshDisplayBackgroundModelProperty, this);
	registerRefresher(displayBackgroundParametersValues, &RemoteLSSSimulation::refreshDisplayBackgroundParametersValuesProperty,
			this);
	registerRefresher(displayBlockedBeamModelName, &RemoteLSSSimulation::refreshDisplayBlockedBeamModelProperty, this);
	registerRefresher(displayBlockedBeamParametersValues,
			&RemoteLSSSimulation::refreshDisplayBlockedBeamParametersValuesProperty, this);
	registerRefresher(useSimulatedData, &RemoteLSSSimulation::refreshUseSimulatedDataProperty, this);

	m_CountSpy = dynamic_cast<utilities::CountSpy*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
			utilities::CountSpy::TYPE));
	if (m_CountSpy != nullptr) {
		m_CountSpy->attach(this);
		registerUpdater(m_CountSpy->statusMessage, &RemoteLSSSimulation::updateCountStatusMessage, this);
		updateCountStatusMessage();
	}
}

/*
 * refreshUseSimulatedDataProperty
 */
void RemoteLSSSimulation::refreshUseSimulatedDataProperty(bool value) throw (CannotSetValue) {
	if (value == true) {
		m_CountSpy->setNoDataRead();
	} else {
		m_CountSpy->clearNoDataRead();
	}
}

/*
 * refreshDisplayModelProperty
 */
void RemoteLSSSimulation::refreshDisplaySampleModelProperty(const std::string& model) throw (CannotSetValue) {
	int32 i = 0;
	for (i = 0; i < model_name.getSize(); ++i) {
		if (model_name.get(i).find(model) != string::npos) {
			break;
		}
	}
	if (i < model_name.getSize()) {
		// load to display
		m_displaySampleModel = i;
		displaySampleModelType.update(model_type.get(i));
		displaySampleParametersNames.resize(parameters.getSize(i));
		displaySampleParametersValues.resize(parameters.getSize(i));
		for (int32 j = 0; j < parameters.getSize(i); ++j) {
			displaySampleParametersNames.update(j, pnames.get(i, j));
			displaySampleParametersValues.update(j, parameters.get(i, j));
		}
		displaySampleNbParameters.update(parameters.getSize(i));
	} else {
		throw CannotSetValue();
	}
}

/*
 * refreshDisplayParametersValuesProperty
 */
void RemoteLSSSimulation::refreshDisplaySampleParametersValuesProperty(int32 i, float64 value) throw (CannotSetValue) {
	parameters.update(m_displaySampleModel, i, value);
}

/*
 * refreshDisplayModelProperty
 */
void RemoteLSSSimulation::refreshDisplayBackgroundModelProperty(const std::string& model) throw (CannotSetValue) {
	int32 i = 0;
	for (i = 0; i < model_name.getSize(); ++i) {
		if (model_name.get(i).find(model) != string::npos) {
			break;
		}
	}
	if (i < model_name.getSize()) {
		// load to display
		m_displayBackgroundModel = i;
		displayBackgroundModelType.update(model_type.get(i));
		displayBackgroundParametersNames.resize(parameters.getSize(i));
		displayBackgroundParametersValues.resize(parameters.getSize(i));
		for (int32 j = 0; j < parameters.getSize(i); ++j) {
			displayBackgroundParametersNames.update(j, pnames.get(i, j));
			displayBackgroundParametersValues.update(j, parameters.get(i, j));
		}
		displayBackgroundNbParameters.update(parameters.getSize(i));
	} else {
		throw CannotSetValue();
	}
}

/*
 * refreshDisplayParametersValuesProperty
 */
void RemoteLSSSimulation::refreshDisplayBackgroundParametersValuesProperty(int32 i, float64 value) throw (CannotSetValue) {
	parameters.update(m_displayBackgroundModel, i, value);
}

/*
 * refreshDisplayModelProperty
 */
void RemoteLSSSimulation::refreshDisplayBlockedBeamModelProperty(const std::string& model) throw (CannotSetValue) {
	int32 i = 0;
	for (i = 0; i < model_name.getSize(); ++i) {
		if (model_name.get(i).find(model) != string::npos) {
			break;
		}
	}
	if (i < model_name.getSize()) {
		// load to display
		m_displayBlockedBeamModel = i;
		displayBlockedBeamModelType.update(model_type.get(i));
		displayBlockedBeamParametersNames.resize(parameters.getSize(i));
		displayBlockedBeamParametersValues.resize(parameters.getSize(i));
		for (int32 j = 0; j < parameters.getSize(i); ++j) {
			displayBlockedBeamParametersNames.update(j, pnames.get(i, j));
			displayBlockedBeamParametersValues.update(j, parameters.get(i, j));
		}
		displayBlockedBeamNbParameters.update(parameters.getSize(i));
	} else {
		throw CannotSetValue();
	}
}

/*
 * refreshDisplayParametersValuesProperty
 */
void RemoteLSSSimulation::refreshDisplayBlockedBeamParametersValuesProperty(int32 i, float64 value) throw (CannotSetValue) {
	parameters.update(m_displayBlockedBeamModel, i, value);
}

/*
 * initApplication
 */
bool RemoteLSSSimulation::initApplication() {

	try {
		// Start the Matlab server
		if (!m_server->isAvailable(10000)) {
			cout << "Matlab server is not available" << endl;
			return false;
		}

		cout << "Matlab server is connected to " << getName() << endl;

		m_matlabApplication = m_server->connect(MATLAB_APPLICATION);

		if (m_matlabApplication->exists() == false) {
			m_matlabApplication = m_server->start(MATLAB_APPLICATION);
			if (!m_matlabApplication->exists()) {
				cout << "No matlab application" << endl;
				return false;
			}
			applicationState = m_matlabApplication->waitFor(application::STARTING);
		}

		// First delete the requester if it exists.
		m_requester.reset(0);

		// Create a new requester.
		m_requester = application::Requester::create(*m_matlabApplication, RESPONDER_NAME);

		if (m_requester.get() == 0) {
			cout << "requester error" << endl;
			return false;
		}

		// Application initialized.
		initialized = true;

		applicationState = m_matlabApplication->waitFor(application::RUNNING);

		cout << "Matlab application initialized" << endl;

		return true;

	} catch (const std::exception & e) {
		// Currently an exception can occur during isAvailable.
		cout << "Unexpected exception during matlab connection: " << e.what() << endl;
	}

	return false;
}

/*
 * test
 */
void RemoteLSSSimulation::test() {
	if (initialized() == false) {
		init();
		if (initialized() == false) {
			return;
		}
	}

	// Serialize the request.
	Request request;
	request.set_type(Request::TEST);

	m_requester->sendBinary(request.SerializeAsString());

	cout << getName() << " sent get data request " << endl;

	// Wait for the response synchronously.
	// Note that responses can be processed asynchronously if the server is able to.
	string data;
	m_requester->receiveBinary(data);
	TestResponse response;
	response.ParseFromString(data);

	cout << "received test response " << response.value() << endl;

}

/*
 * start
 */
void RemoteLSSSimulation::start() {
	commandProgression = 0;
	commandProgression = 100;
}

/*
 * getData
 */
void RemoteLSSSimulation::getData() {
	if (initialized() == false) {
		init();
		if (initialized() == false) {
			return;
		}
	}

	// Serialize the request.
	Request request;
	request.set_type(Request::GET_DATA);

	// Serialize the request.
	GetDataRequest params;

	// 'instrument'       :  'd11','d22' or 'd33'
	acquisition::ExperimentData* title =
			dynamic_cast<acquisition::ExperimentData*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
					acquisition::ExperimentData::TYPE));
	if (title == nullptr) {
		log(Level::s_Error) << " missing controller Title in this instrument" << endlog;
		commandStatus.setError();
		commandProgression = 100;
		return;
	}
	params.set_instrument(title->instrumentName());

	//	%'measurement_time' :  600  [seconds]
	if (m_CountSpy == nullptr) {
		log(Level::s_Error) << " missing controller Count Spy in this instrument" << endlog;
		commandStatus.setError();
		commandProgression = 100;
		return;
	}
	params.set_measurement_time(m_CountSpy->time());

	//	%'source_size'      :  e.g. rectangular [30e-3, 30e-3] [m] or circular [30e-3] [m]
	if (title->instrumentName() == "D11") {
		lss::CollimationProperties* collimation =
				dynamic_cast<lss::CollimationProperties*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
						D11_COLLIMATION_TYPE));
		if (collimation == nullptr) {
			log(Level::s_Error) << " missing controller Collimation in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
		typedef tokenizer<char_separator<char> > tokenizer;
		char_separator<char> sep(" ()x-");
		tokenizer tok1(collimation->sourceAperture(), sep);
		tokenizer::iterator iter = tok1.begin();
		vector<float64> values;
		try {
			if (iter != tok1.end()) {
				++iter;
				if (iter != tok1.end()) {
					// get first dimension
					values.push_back(lexical_cast<float64>(*iter) / 1000.);
					++iter;
					if (iter != tok1.end()) {
						// get second dimension
						values.push_back(lexical_cast<float64>(*iter) / 1000.);
					}
				}
				vector<float64>::iterator valuesiter = values.begin();
				while (valuesiter != values.end()) {
					params.add_source_size(*valuesiter);
					++valuesiter;
				}
			}
		} catch (boost::bad_lexical_cast& e) {
			log(Level::s_Error) << " can't get source size from d11 Collimation setting controller" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
	} else if (title->instrumentName() == "D22") {
		lss::CollimationProperties* collimation =
				dynamic_cast<lss::CollimationProperties*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
						D22_COLLIMATION_TYPE));
		if (collimation == nullptr) {
			log(Level::s_Error) << " missing controller Collimation in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
		for (int32 i = 0; i < collimation->selectedApertInfo.getSize(); ++i) {
			params.add_source_size(collimation->selectedApertInfo()[i] / 1000.);
		}
	} else if (title->instrumentName() == "D33") {
		lss::CollimationProperties* collimation =
				dynamic_cast<lss::CollimationProperties*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
						D33_COLLIMATION_TYPE));
		if (collimation == nullptr) {
			log(Level::s_Error) << " missing controller Collimation in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
		for (int32 i = 0; i < collimation->ap_size.getSize(); ++i) {
			params.add_source_size(collimation->ap_size()[i] / 1000.);
		}
	}

	//	%'col' .            :  e.g. 12.8, 10.3, 7.8, 5.3, or 2.8 [m]
	if (title->instrumentName() == "D11") {
		lss::CollimationProperties* collimation =
				dynamic_cast<lss::CollimationProperties*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
						D11_COLLIMATION_TYPE));
		if (collimation == nullptr) {
			log(Level::s_Error) << " missing controller Collimation in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
		params.set_col(collimation->collimation());
	} else if (title->instrumentName() == "D22") {
		lss::CollimationProperties* collimation =
				dynamic_cast<lss::CollimationProperties*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
						D22_COLLIMATION_TYPE));
		if (collimation == nullptr) {
			log(Level::s_Error) << " missing controller Collimation in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
		params.set_col(collimation->selectedColSetup());
	} else if (title->instrumentName() == "D33") {
		lss::CollimationProperties* collimation =
				dynamic_cast<lss::CollimationProperties*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
						D33_COLLIMATION_TYPE));
		if (collimation == nullptr) {
			log(Level::s_Error) << " missing controller Collimation in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
		params.set_col(collimation->collimation());
	}

//	%'sample_size'      :  e.g. rectangular [10e-3, 10e-3] [m] or circular [10e-3] [m]
	utilities::DataParams* dataparams =
			dynamic_cast<utilities::DataParams*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
					utilities::DataParams::TYPE));
	if (dataparams == nullptr) {
		log(Level::s_Error) << " missing controller Data Params in this instrument" << endlog;
		commandStatus.setError();
		commandProgression = 100;
		return;
	}
	if (title->instrumentName() == "D11") {
		for (int32 i = 0; i < dataparams->nbGroups(); ++i) {
			for (int32 j = 0; j < dataparams->groupSize.get(i); ++j) {
				if (dataparams->valueName.get(i, j).find("Width or Diameter") != string::npos) {
					params.add_sample_size(dataparams->valueFloat.get(i, j) / 1000.);
				} else if (dataparams->valueName.get(i, j).find("Height or Zero") != string::npos) {
					if (dataparams->valueFloat.get(i, j) != 0) {
						params.add_sample_size(dataparams->valueFloat.get(i, j) / 1000.);
					}
				}
			}
		}
	} else if (title->instrumentName() == "D22") {
		for (int32 i = 0; i < dataparams->nbGroups(); ++i) {
			for (int32 j = 0; j < dataparams->groupSize.get(i); ++j) {
				if (dataparams->valueName.get(i, j).find("Horizontal or Diameter") != string::npos) {
					params.add_sample_size(dataparams->valueFloat.get(i, j) / 1000.);
				} else if (dataparams->valueName.get(i, j).find("Vertical or Zero") != string::npos) {
					if (dataparams->valueFloat.get(i, j) != 0) {
						params.add_sample_size(dataparams->valueFloat.get(i, j) / 1000.);
					}
				}
			}
		}
	} else if (title->instrumentName() == "D33") {
		for (int32 i = 0; i < dataparams->nbGroups(); ++i) {
			for (int32 j = 0; j < dataparams->groupSize.get(i); ++j) {
				if (dataparams->valueName.get(i, j).find("Horizontal or Diameter") != string::npos) {
					params.add_sample_size(dataparams->valueFloat.get(i, j) / 1000.);
				} else if (dataparams->valueName.get(i, j).find("Vertical or Zero") != string::npos) {
					if (dataparams->valueFloat.get(i, j) != 0) {
						params.add_sample_size(dataparams->valueFloat.get(i, j) / 1000.);
					}
				}
			}
		}
	}

//	%'wav'              :  6 [angs] (monochromatic) or [1,2,3,4,5,6,....] (array of wavelength bin-centres for tof)
	lss::LSSWavelength* wavelength =
			dynamic_cast<lss::LSSWavelength*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
					lss::LSSWavelength::TYPE));
	if (wavelength == nullptr) {
		log(Level::s_Error) << " missing controller Wavelength in this instrument" << endlog;
		commandStatus.setError();
		commandProgression = 100;
		return;
	}
	params.set_wav(wavelength->wavelength());

//	%'det'              :  10 [m] (detector distance for single detector instrument, D11 & D22
	if (title->instrumentName() != "D33") {
		try {
			axis::AxisController* det =
					dynamic_cast<axis::AxisController*>(InstrumentManager::getInstance()->getExperimentController("Det"));
			params.set_det(det->position());
		} catch (InstrumentManager::ControllerNotFound& e) {
			log(Level::s_Error) << " missing controller Det in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
	}

//	%'det1'             :  2 [m] (front detector d33)
	if (title->instrumentName() == "D33") {
		try {
			axis::AxisController* det1 =
					dynamic_cast<axis::AxisController*>(InstrumentManager::getInstance()->getExperimentController("Det1"));
			params.set_det1(det1->position());
		} catch (InstrumentManager::ControllerNotFound& e) {
			log(Level::s_Error) << " missing controller Det1 in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
	}

//	%'det2'             :  8 [m] (rear detector d33)
	if (title->instrumentName() == "D33") {
		try {
			axis::AxisController* det2 =
					dynamic_cast<axis::AxisController*>(InstrumentManager::getInstance()->getExperimentController("Det2"));
			params.set_det2(det2->position());
		} catch (InstrumentManager::ControllerNotFound& e) {
			log(Level::s_Error) << " missing controller Det2 in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
	}

//	%'bx'               :  0 [mm] (beamstop x position)
	try {
		axis::AxisController* bx =
				dynamic_cast<axis::AxisController*>(InstrumentManager::getInstance()->getExperimentController("Bx"));
		params.set_bx(bx->position());
	} catch (InstrumentManager::ControllerNotFound& e) {
		log(Level::s_Error) << " missing controller Bx in this instrument" << endlog;
		commandStatus.setError();
		commandProgression = 100;
		return;
	}

//	%'by' .             :  0 [mm] (beamstop y position)
	try {
		axis::AxisController* by =
				dynamic_cast<axis::AxisController*>(InstrumentManager::getInstance()->getExperimentController("By"));
		params.set_by(by->position());
	} catch (InstrumentManager::ControllerNotFound& e) {
		log(Level::s_Error) << " missing controller By in this instrument" << endlog;
		commandStatus.setError();
		commandProgression = 100;
		return;
	}

//	%'attenuator'       :  1,  100, 1000 (actual attenuation value in use)
	if (title->instrumentName() == "D11") {
		try {
			stateio::AttenuatorController* attenuator =
					dynamic_cast<stateio::AttenuatorController*>(InstrumentManager::getInstance()->getExperimentController(
							"Attenuator"));
			params.set_attenuator(attenuator->attenuationCoefficient());
		} catch (InstrumentManager::ControllerNotFound& e) {
			log(Level::s_Error) << " missing controller Attenuator in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
	} else if (title->instrumentName() == "D22") {
		try {
			stateaxis::StateAxisController* attenuator =
					dynamic_cast<stateaxis::StateAxisController*>(InstrumentManager::getInstance()->getExperimentController(
							"Attenuator"));
			params.set_attenuator(lexical_cast<float64>(attenuator->actualStateValue()));
		} catch (InstrumentManager::ControllerNotFound& e) {
			log(Level::s_Error) << " missing controller Attenuator in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
	} else if (title->instrumentName() == "D33") {
		try {
			stateaxis::StateAxisController* attenuator =
					dynamic_cast<stateaxis::StateAxisController*>(InstrumentManager::getInstance()->getExperimentController(
							"Attenuator"));
			float64 val = lexical_cast<float64>(attenuator->actualStateValue());
			if (val == 0) {
				val = 1.;
			}
			params.set_attenuator(val);
		} catch (InstrumentManager::ControllerNotFound& e) {
			log(Level::s_Error) << " missing controller Attenuator in this instrument" << endlog;
			commandStatus.setError();
			commandProgression = 100;
			return;
		}
	}
//	%'thickness'        :  0.1 [cm] (sample thickness - applies only to the sample scatter model, not the background and blocked models
//	utilities::DataParams* dataparams =
//			dynamic_cast<utilities::DataParams*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
//					utilities::DataParams::TYPE));
//	if (dataparams == nullptr) {
//		log(Level::s_Error) << " missing controller Data Params in this instrument" << endlog;
//		commandStatus.setError();
//		commandProgression = 100;
//		return;
//	}
	if (title->instrumentName() == "D11") {
		for (int32 i = 0; i < dataparams->nbGroups(); ++i) {
			for (int32 j = 0; j < dataparams->groupSize.get(i); ++j) {
				if (dataparams->valueName.get(i, j).find("Thickness") != string::npos) {
					params.set_thickness(dataparams->valueFloat.get(i, j));
				}
			}
		}
	} else if (title->instrumentName() == "D22") {
		for (int32 i = 0; i < dataparams->nbGroups(); ++i) {
			for (int32 j = 0; j < dataparams->groupSize.get(i); ++j) {
				if (dataparams->valueName.get(i, j).find("Thickness") != string::npos) {
					params.set_thickness(dataparams->valueFloat.get(i, j));
				}
			}
		}
	} else if (title->instrumentName() == "D33") {
		for (int32 i = 0; i < dataparams->nbGroups(); ++i) {
			for (int32 j = 0; j < dataparams->groupSize.get(i); ++j) {
				if (dataparams->valueName.get(i, j).find("Thickness") != string::npos) {
					params.set_thickness(dataparams->valueFloat.get(i, j));
				}
			}
		}
	}

	// Sample model
	ModelParam* samplemodel = params.mutable_sample_model();
	samplemodel->set_name(displaySampleModelName());
	for (int32 i = 0; i < parameters.getSize(m_displaySampleModel); ++i) {
		samplemodel->add_structname(structname.get(m_displaySampleModel, i));
		samplemodel->add_parameters(parameters.get(m_displaySampleModel, i));
	}

	// Background
	ModelParam* backgroundmodel = params.mutable_background_model();
	backgroundmodel->set_name(displayBackgroundModelName());
	for (int32 i = 0; i < parameters.getSize(m_displayBackgroundModel); ++i) {
		backgroundmodel->add_structname(structname.get(m_displayBackgroundModel, i));
		backgroundmodel->add_parameters(parameters.get(m_displayBackgroundModel, i));
	}

	// Blocked model
	ModelParam* blockedbeammodel = params.mutable_blockedbeam_model();
	blockedbeammodel->set_name(displayBlockedBeamModelName());
	for (int32 i = 0; i < parameters.getSize(m_displayBlockedBeamModel); ++i) {
		blockedbeammodel->add_structname(structname.get(m_displayBlockedBeamModel, i));
		blockedbeammodel->add_parameters(parameters.get(m_displayBlockedBeamModel, i));
	}

//	cout << "instrument = " << params.instrument() << endl;
//	cout << "measurement_time = " << params.measurement_time() << endl;
//	cout << "source_size = ";
//	for (int32 i = 0; i < params.source_size_size(); ++i) {
//		cout << params.source_size((int32_t) i) << " , ";
//	}
//	cout << endl;
//	cout << "col = " << params.col() << endl;
//	cout << "sample_size = ";
//	for (int32 i = 0; i < params.sample_size_size(); ++i) {
//		cout << params.sample_size((int32_t) i) << " , ";
//	}
//	cout << endl;
//	cout << "wav = " << params.wav() << endl;
//	cout << "det = " << params.det() << endl;
//	cout << "det1 = " << params.det1() << endl;
//	cout << "det2 = " << params.det2() << endl;
//	cout << "bx = " << params.bx() << endl;
//	cout << "by = " << params.by() << endl;
//	cout << "attenuator = " << params.attenuator() << endl;
//	cout << "thickness = " << params.thickness() << endl;
//    cout << "sample model name = " << params.sample_model().name() << endl;
//    cout << "background model name = " << params.background_model().name() << endl;
//    cout << "blockedbeam model name = " << params.blockedbeam_model().name() << endl;

//	cout << getName() << " sent get data request " << endl;

	m_requester->sendTwoBinaryParts(request.SerializeAsString(), params.SerializeAsString());

// Wait for the response synchronously.
// Note that responses can be processed asynchronously if the server is able to.
	string datastring;
	m_requester->receiveBinary(datastring);
	m_DataResponse.Clear();
	m_DataResponse.ParseFromString(datastring);
	if (m_DataResponse.error().empty() == false) {
		log(Level::s_Error) << " matlab error: " << m_DataResponse.error() << endlog;
	}
}

/*
 * init
 */
void RemoteLSSSimulation::init() {

	cout << "RemoteLSSSimulation::init" << endl;
// Do not initialize if it is already done
	if (initialized()) {
		return;
	}

//	if (isEnabled() == false) {
//		return;
//	}

	if (!initApplication()) {
		return;
	}

// Serialize the request.
	Request request;
	request.set_type(Request::GET_MODELS);

// Send the message.
	m_requester->sendBinary(request.SerializeAsString());

	cout << getName() << " sent get models request " << endl;

// Wait for the response synchronously.
// Note that responses can be processed asynchronously if the server is able to.
	string data;
	m_requester->receiveBinary(data);

	ModelsResponse response;
	response.ParseFromString(data);

	cout << "received response " << response.models().size() << endl;
	if (response.error().empty() == false) {
		log(Level::s_Error) << " matlab error: " << response.error() << endlog;
	} else {
		resizeAll(response.models().size());
		for (int32 i = 0; i < response.models().size(); ++i) {
			model_name.update(i, response.models(i).name());
			model_type.update(i, response.models(i).type());
			if (model_type.get(i) == SAMPLE_TYPE) {
				sampleModelName.resize(sampleModelName.getSize() + 1);
				sampleModelName.update(sampleModelName.getSize() - 1, model_name.get(i));
			} else if (model_type.get(i) == BACKGROUND_TYPE) {
				backgroundModelName.resize(backgroundModelName.getSize() + 1);
				backgroundModelName.update(backgroundModelName.getSize() - 1, model_name.get(i));
			} else if (model_type.get(i) == BLOCKEDBEAM_TYPE) {
				blockedBeamModelName.resize(blockedBeamModelName.getSize() + 1);
				blockedBeamModelName.update(blockedBeamModelName.getSize() - 1, model_name.get(i));
			}
			model_number.update(i, response.models(i).model_number());
			pnames.resize(i, response.models(i).pnames().size());
			for (int32 j = 0; j < response.models(i).pnames().size(); ++j) {
				pnames.update(i, j, response.models(i).pnames(j));
			}
			structname.resize(i, response.models(i).structname().size());
			for (int32 j = 0; j < response.models(i).structname().size(); ++j) {
				structname.update(i, j, response.models(i).structname(j));
			}
			fn_eval.update(i, response.models(i).fn_eval());
			parameters.resize(i, response.models(i).parameters().size());
			for (int32 j = 0; j < response.models(i).parameters().size(); ++j) {
				parameters.update(i, j, response.models(i).parameters(j));
			}
		}

		nbSampleModels.update(sampleModelName.getSize());
		nbBackgroundModels.update(backgroundModelName.getSize());
		nbBlockedBeamModels.update(blockedBeamModelName.getSize());

		displaySampleModelName.setEnumeratedLabels(sampleModelName);
		displaySampleModelName.setEnumeratedValues(sampleModelName);
		displaySampleModelName = sampleModelName.get(0);

		displayBackgroundModelName.setEnumeratedLabels(backgroundModelName);
		displayBackgroundModelName.setEnumeratedValues(backgroundModelName);
		displayBackgroundModelName = backgroundModelName.get(0);

		displayBlockedBeamModelName.setEnumeratedLabels(blockedBeamModelName);
		displayBlockedBeamModelName.setEnumeratedValues(blockedBeamModelName);
		displayBlockedBeamModelName = blockedBeamModelName.get(0);

		// Do not initialise if the controller is disabled.
		initialized = true;
	}

}

/*
 * resizeAll
 */
void RemoteLSSSimulation::resizeAll(int32 size) {
	model_name.resize(size);
	model_type.resize(size);
	model_number.resize(size);
	pnames.resize(size);
	structname.resize(size);
	fn_eval.resize(size);
	parameters.resize(size);
}

/*
 * updateCountStatusMessage
 */
void RemoteLSSSimulation::updateCountStatusMessage() {
	if (useSimulatedData() == true) {
		string statusmes = m_CountSpy->statusMessage();

		if (statusmes == acquisition::Count::WRITING_DATA_MESSAGE) {
			acquisition::ExperimentData* title =
					dynamic_cast<acquisition::ExperimentData*>(InstrumentManager::getInstance()->getFirstExperimentControllerByType(
							acquisition::ExperimentData::TYPE));
			if (title == nullptr) {
				log(Level::s_Error) << " missing controller Title in this instrument" << endlog;
				commandStatus.setError();
				commandProgression = 100;
				return;
			}
			// Re-write data to detector controller
			acquisition::Count* count = nullptr;
			try {
				count = dynamic_cast<acquisition::Count*>(InstrumentManager::getInstance()->getExperimentController(
						m_CountSpy->countControllerName()));
			} catch (InstrumentManager::ControllerNotFound& e) {
				log(Level::s_Error) << " missing controller " << m_CountSpy->countControllerName() << " in this instrument"
						<< endlog;
				commandStatus.setError();
				commandProgression = 100;
				return;
			}

			getData();
			cout << "received monitor " << m_DataResponse.monitor() << endl;
			cout << "received response " << m_DataResponse.datas_size() << endl;
			if (count->nMonitors() > 0) {
				if (count->monitors[0]->data.getSize() > 0) {
					count->monitors[0]->data()[0] = m_DataResponse.monitor();
					count->monitors[0]->detSum.update(m_DataResponse.monitor());
				}
			}

			if (m_DataResponse.datas_size() >= 1) {
				Data data = m_DataResponse.datas(0);
				int32* detdata = count->masterDetector->data();
				float64 sum = 0.;
				if ((data.sizex() * data.sizey()) == count->masterDetector->data.getSize()) {
					for (int32 j = 0; j < data.data_size(); ++j) {
						detdata[j] = data.data(j);
						sum += detdata[j];
					}
				}
				count->masterDetector->detSum.update(sum);
			}
			if (title->instrumentName() == "D33") {
				if (m_DataResponse.datas_size() >= 2) {
					try {
						acquisition::DetectorElement* det =
								dynamic_cast<acquisition::DetectorElement*>(InstrumentManager::getInstance()->getExperimentController(
										"MultiDetector2"));
						Data data = m_DataResponse.datas(1);
						int32* detdata = det->data();
						float64 sum = 0.;
						if ((data.sizex() * data.sizey()) == det->data.getSize()) {
							for (int32 j = 0; j < data.data_size(); ++j) {
								detdata[j] = data.data(j);
								sum += detdata[j];
							}
						}
						det->detSum.update(sum);
					} catch (InstrumentManager::ControllerNotFound& e) {
						log(Level::s_Error) << " missing controller MultiDetector2 in this instrument" << endlog;
						commandStatus.setError();
						commandProgression = 100;
						return;
					}
				}
				if (m_DataResponse.datas_size() >= 3) {
					try {
						acquisition::DetectorElement* det =
								dynamic_cast<acquisition::DetectorElement*>(InstrumentManager::getInstance()->getExperimentController(
										"MultiDetector3"));
						Data data = m_DataResponse.datas(2);
						int32* detdata = det->data();
						float64 sum = 0.;
						if ((data.sizex() * data.sizey()) == det->data.getSize()) {
							for (int32 j = 0; j < data.data_size(); ++j) {
								detdata[j] = data.data(j);
								sum += detdata[j];
							}
						}
						det->detSum.update(sum);
					} catch (InstrumentManager::ControllerNotFound& e) {
						log(Level::s_Error) << " missing controller MultiDetector3 in this instrument" << endlog;
						commandStatus.setError();
						commandProgression = 100;
						return;
					}
				}
				if (m_DataResponse.datas_size() >= 4) {
					try {
						acquisition::DetectorElement* det =
								dynamic_cast<acquisition::DetectorElement*>(InstrumentManager::getInstance()->getExperimentController(
										"MultiDetector4"));
						Data data = m_DataResponse.datas(3);
						int32* detdata = det->data();
						float64 sum = 0.;
						if ((data.sizex() * data.sizey()) == det->data.getSize()) {
							for (int32 j = 0; j < data.data_size(); ++j) {
								detdata[j] = data.data(j);
								sum += detdata[j];
							}
						}
						det->detSum.update(sum);
					} catch (InstrumentManager::ControllerNotFound& e) {
						log(Level::s_Error) << " missing controller MultiDetector4 in this instrument" << endlog;
						commandStatus.setError();
						commandProgression = 100;
						return;
					}
				}
				if (m_DataResponse.datas_size() >= 5) {
					try {
						acquisition::DetectorElement* det =
								dynamic_cast<acquisition::DetectorElement*>(InstrumentManager::getInstance()->getExperimentController(
										"MultiDetector5"));
						Data data = m_DataResponse.datas(4);
						int32* detdata = det->data();
						float64 sum = 0.;
						if ((data.sizex() * data.sizey()) == det->data.getSize()) {
							for (int32 j = 0; j < data.data_size(); ++j) {
								detdata[j] = data.data(j);
								sum += detdata[j];
							}
						}
						det->detSum.update(sum);
					} catch (InstrumentManager::ControllerNotFound& e) {
						log(Level::s_Error) << " missing controller MultiDetector5 in this instrument" << endlog;
						commandStatus.setError();
						commandProgression = 100;
						return;
					}
				}

			}
		}
	}
}

}
