/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "drivers/gigecam/photonics/RealPhotonicsDriver.h"

#include <boost/lexical_cast.hpp>

#include "CImg.h"

#include "drivers/gigecam/photonics/PhotonicsDef.h"
#include "drivers/gigecam/photonics/PhotonicsDriver.h"
using namespace cimg_library;

using namespace std;

///
/// \namespace photonics
/// \brief  camera namespace
///
namespace photonics {

/*
 * Constructor
 */
RealPhotonicsDriver::RealPhotonicsDriver(PhotonicsDriver* owner) :
				PhotonicsState(owner) {
	m_return_code = 0;
	m_finish_console_threads = false;
	m_command_status = false;
}

/*
 * Destructor
 */
RealPhotonicsDriver::~RealPhotonicsDriver() {
	if (m_return_code == 0) {
		// set the cancel flag of status thread
		m_finish_console_threads = true;

		// free all BGAPI resources and worker threads
		m_camera_handler.StopCameras();
		m_camera_handler.Stop();
	}
}

//string RealPhotonicsDriver::getConfigFile(BGAPI2::Device* device) {
//	std::string confgFile;
//	//	BGAPI2::Node * pNode = NULL;
//	//	BGAPI2::NodeMap * pNodeTree = NULL;
//	//	if (device != 0) {
//	//		if (device->IsOpen()) {
//	//
//	//			pNodeTree = device->GetRemoteNodeTree();
//	//
//	//			for (bo_uint64 i = 0; i < device->GetRemoteNodeTree()->GetNodeCount(); i++) {
//	//				pNode = pNodeTree->GetNodeByIndex(i);
//	//				printNodeRecursive(pNode, 0);
//	//			}
//	//
//	//			confgFile = device->GetRemoteConfigurationFile();
//	//		}
//	//	}
//	//	cout << "confgFile: " << confgFile << endl;
//	return confgFile;
//}

/*
 * init
 */
void RealPhotonicsDriver::init() {
	// configure the camera handler to work the double buffer handler
	// at this point it is possible to implement other buffer exchange strategies
	m_camera_handler.Start(m_double_buffer);

	// check, if everything is correct initialized and at least one camera was found
	if (m_camera_handler.IsInitOK()) {
		std::cout << "all connected cameras are successfully initialized and opened" << std::endl;
	} else {
		do {
			std::cout << m_camera_handler.LoggingString() << std::endl;
		} while (!m_camera_handler.LoggingListEmpty());
		std::cout << "problem while initializing the BGAPI, program finished." << std::endl;
		m_return_code = 1;
	}
	if (m_camera_handler.NumberOfCameras() == 0) {
		std::cout << "no cameras connected, program finished." << std::endl;
		m_return_code = 1;
	}

	// print out some logging information, if there are some
	// and abort the program, if something goes wrong
	do {
		std::cout << m_camera_handler.LoggingString() << std::endl;
	} while (!m_camera_handler.LoggingListEmpty());

	if (m_return_code == 1) {
		m_camera_handler.Stop();
		throw DeviceDriver::TimeOut(owner()->getType(), owner()->getName(), owner(),
				"Problem while initializing the BGAPI, program finished");
	}

	// start and configure all connected cameras
	m_camera_handler.StartCameras();

	m_return_code = 0;
}

//void RealPhotonicsDriver::InitCurrentBufferStruct(CCameraHandler *camera_handler, std::list<std::thread::id> thread_id_list) {
//    const std::vector<CCameraControl*>* camera_vector = camera_handler->GetCameraControlVector();
//    for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
//        camera_iter != camera_vector->end();
//        camera_iter++) {
//        for (std::list<std::thread::id>::iterator iter = thread_id_list.begin(); iter != thread_id_list.end(); iter++) {
//            m_current_buffer[&(*camera_iter)->buffer_handler_].new_data[*iter] = true;
//        }
//    }
//}
//
//// the StatusThreadRoutine is left and the status thread finished, if the finish flag was set
//// the StatusThreadRoutine starts working if the status command was activated from within the main loop
//// the StatusThreadRoutine works with polling to detect an activated status command
//void RealPhotonicsDriver::StatusThreadRoutine(CCameraHandler* camera_handler) {
//	// abort condition of the thread routine
//	while (!m_finish_console_threads) {
//		// check if the user activated the status command
//		if (m_command_status) {
//			const std::vector<CCameraControl*>* camera_vector = camera_handler->GetCameraControlVector();
//			for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
//					camera_iter != camera_vector->end(); camera_iter++) {
//				// perform logging output
//				PrintLoggingInformation(*camera_iter);
//				// perform status output
//				PrintCameraStatus(*camera_iter);
//			}
//			// status command is now finished
//			m_command_status = false;
//		} else {
//			// just wait until status command become active
//			std::this_thread::sleep_for(std::chrono::milliseconds(10));
//		}
//	}
//}
//
//// the PrintCameraStatus receive one camera buffer with buffer information from the passed camera
//// and print this information to the console output
//// the PrintCameraStatus use the synchronisation functions for a synchronized buffer access
//// the PrintCameraStatus works only once on the same camera buffer on a camera (see HasDoubleBufferHandlerNewData)
//void RealPhotonicsDriver::PrintCameraStatus(CCameraControl * camera_control) {
//	std::cout << "status of camera " << camera_control->camera_pointer_->GetDisplayName() << std::endl;
//	// check for a new camera buffer
//	if (HasDoubleBufferHandlerNewData(&camera_control->buffer_handler_)) {
//		if (camera_control->buffer_handler_.HasNewData()) {
//			// receive a camera buffer
//			BGAPI2::Buffer * buffer = camera_control->buffer_handler_.PullBuffer();
//			if (buffer) {
//				BufferInformation * buffer_info = reinterpret_cast<BufferInformation*>(buffer->GetUserObj());
//			//	if (!buffer->GetIsIncomplete()) {
//					if (camera_control->capture_active_) {
//						std::cout << "capturing is active" << std::endl;
//					} else {
//						std::cout << "capturing is inactive" << std::endl;
//					}
//					std::cout << "number of captured images: " << camera_control->number_of_captured_images_ << std::endl;
//					std::cout << "number of incomplete frames: " << buffer_info->number_of_incomplete_images << std::endl;
//					std::cout << "resend requests: " << buffer_info->resend_requests << " single: "
//							<< buffer_info->resend_requests_single << " range: " << buffer_info->resend_requests_range << std::endl;
//					if (buffer_info->frameid > 0) {
//						std::cout << "last processed image id: " << buffer_info->frameid << std::endl;
//					}
//				//} else {
//				//	std::cout << "buffer status skipped, current buffer was not completely received." << std::endl
//				//			<< "captured images: " << camera_control->number_of_captured_images_ << std::endl;
//				//}
//				// return the camera buffer
//				FreeBufferToDoubleBufferHandler(&camera_control->buffer_handler_, buffer);
//			} else {
//				std::cout << "no new buffer captured: " << camera_control->number_of_captured_images_ << std::endl;
//			}
//		} else {
//			std::cout << "no new buffer available" << std::endl << "captured images: " << camera_control->number_of_captured_images_
//					<< std::endl;
//		}
//		std::stringstream outputstream;
//		{
//			std::lock_guard<std::mutex> lock(camera_control->feature_command_message_list_lock_);
//			for (std::list<std::string>::iterator msg_iter = camera_control->feature_command_message_list_.begin();
//					msg_iter != camera_control->feature_command_message_list_.end(); msg_iter++) {
//				outputstream << (*msg_iter) << std::endl;
//			}
//		}
//		std::cout << outputstream.str() << std::endl;
//		std::cout << std::endl << std::endl;
//	}
//}
//// this function sets the new data flag for all connected cameras to true, if the buffer_ pointer is nullptr,
//// that means if all application threads have returned their camera buffer
//// and a new camera buffer is served by the buffer handler
//bool RealPhotonicsDriver::HasDoubleBufferHandlerNewData(CDoubleBufferHandler* buffer_handler) {
//	std::lock_guard<std::mutex> lock(m_current_buffer[buffer_handler].lock_);
//	if (m_current_buffer[buffer_handler].buffer_ == nullptr) {
//		if (buffer_handler->HasNewData()) {
//			std::map<std::thread::id, std::atomic<bool>>::iterator it = m_current_buffer[buffer_handler].new_data.begin();
//			for( ; it != m_current_buffer[buffer_handler].new_data.end(); it++) {
//				(*it).second = true;
//			}
//		}
//	}
//	return m_current_buffer[buffer_handler].new_data[std::this_thread::get_id()];
//}
//
//void RealPhotonicsDriver::PrintLoggingInformation(CCameraControl * camera_control) {
//	while (!camera_control->LoggingListEmpty()) {
//		std::cout << camera_control->LoggingString() << std::endl;
//	};
//}
//
//// the function PullBufferFromDoubleBufferHandler uses a mutex to synchronize
//// multiple calls to this function and to function FreeBufferToDoubleBufferHandler
//// the mutex is entered, when this function is called and when the lock is not
//// entered from FreeBufferToDoubleBufferHandler, in this case the current thread
//// stops execution on the mutex object until FreeBufferToDoubleBufferHandler was left
//BGAPI2::Buffer* RealPhotonicsDriver::PullBufferFromDoubleBufferHandler(CDoubleBufferHandler* buffer_handler) {
//	std::lock_guard<std::mutex> lock(m_current_buffer[buffer_handler].lock_);
//	if (m_current_buffer[buffer_handler].buffer_ == nullptr) {
//		m_current_buffer[buffer_handler].buffer_ = buffer_handler->PullBuffer();
//	}
//	if (m_current_buffer[buffer_handler].buffer_ != nullptr) {
//		m_current_buffer[buffer_handler].ref_counter_++;
//	}
//	m_current_buffer[buffer_handler].new_data[std::this_thread::get_id()] = false;
//	return m_current_buffer[buffer_handler].buffer_;
//}
//
//
//// vise versa to function PullBufferFromDoubleBufferHandler
//void RealPhotonicsDriver::FreeBufferToDoubleBufferHandler(CDoubleBufferHandler* buffer_handler, BGAPI2::Buffer* buffer) {
//	std::lock_guard<std::mutex> lock(m_current_buffer[buffer_handler].lock_);
//	if (buffer == m_current_buffer[buffer_handler].buffer_) {
//		if (m_current_buffer[buffer_handler].ref_counter_ > 0) {
//			m_current_buffer[buffer_handler].ref_counter_--;
//		}
//		if (m_current_buffer[buffer_handler].ref_counter_ == 0) {
//			buffer_handler->FreeBuffer(buffer);
//			m_current_buffer[buffer_handler].buffer_ = nullptr;
//		}
//	}
//}

/*
 * clear
 */
void RealPhotonicsDriver::clear() {
}

/*
 * writeParam
 */
void RealPhotonicsDriver::writeParam() {
	ostringstream buf_conf;
	std::size_t found;

	string err;

}

/*
 * synchroniseRead
 */
void RealPhotonicsDriver::synchroniseRead() {

	vector<uint16*> datas;
	datas.resize(m_camera_handler.NumberOfCameras());
	map<CDoubleBufferHandler*, BGAPI2::Buffer *> buffers;
	std::vector<uint32> widths;
	std::vector<uint32> heights;
	widths.resize(m_camera_handler.NumberOfCameras());
	heights.resize(m_camera_handler.NumberOfCameras());

	const std::vector<CCameraControl*>* camera_vector = m_camera_handler.GetCameraControlVector();
	// iterate over all cameras and print the status messages to console
	for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin(); camera_iter != camera_vector->end();
			camera_iter++) {
		std::stringstream result_stream;
		CCameraControl* camera_control = *camera_iter;
	//	if (camera_control->buffer_handler_.HasNewData()) {
			//        if (HasDoubleBufferHandlerNewData(&camera_control->buffer_handler_)) {
			BGAPI2::Buffer * buffer = camera_control->buffer_handler_.PullBuffer();
			if (buffer) {
				BufferInformation * buffer_info = reinterpret_cast<BufferInformation*>(buffer->GetUserObj());
			//	if (!buffer->GetIsIncomplete()) {
					unsigned char* image_buffer = static_cast<unsigned char*>(buffer->GetMemPtr());
					int32 camindice = boost::lexical_cast<int32>((*camera_iter)->camera_pointer_->GetDisplayName());
					cout << "get buffer for camera: " << camindice << endl;

					datas[camindice - 1] = static_cast<uint16*>(buffer->GetMemPtr());
					buffers[&camera_control->buffer_handler_] = buffer;
					widths[camindice - 1] = static_cast<int>(buffer->GetWidth());
					heights[camindice - 1] = static_cast<int>(buffer->GetHeight());
			//	} else {
			//		cout << "current buffer was not completely received!" << endl;
			//		owner()->m_SynchroniseReadTerminated = true;
			//		throw DeviceDriver::TimeOut(owner()->getType(), owner()->getName(), owner(),
				//			"Current buffer was not completely received!");
			//	}
			} else {
				cout << "No image for calculation captured!" << endl;
				owner()->m_SynchroniseReadTerminated = true;
				throw DeviceDriver::TimeOut(owner()->getType(), owner()->getName(), owner(),
						"No image for calculation captured!");
			}
//} else {
//			cout << "no new buffer available!" << endl;
//			owner()->m_SynchroniseReadTerminated = true;
		//	throw DeviceDriver::TimeOut(owner()->getType(), owner()->getName(), owner(), "No new buffer available!");
//		}

	}
	int32* data = owner()->data();
	if (data == nullptr) {
		return;
	}
	int32 dataSize = owner()->dataSize();
	float64 sum = 0;
	CImg<uint16> total;
	for (int32 row = 0; row < owner()->nbRows(); ++row) {
		CImg<uint16> subtotal;
		for (int32 col = 0; col < owner()->nbCols(); ++col) {
			uint16* tmp = datas[col + row * owner()->nbCols()];
			CImg<uint16> cam(tmp, widths[col + row * owner()->nbCols()], heights[col + row * owner()->nbCols()]);
			cout << cam._width << endl;
			cout << cam._height << endl;
			cam.permute_axes("yx");
			cout << cam._width << endl;
			cout << cam._height << endl;
			subtotal.append(cam, 'y');
			cout << subtotal._width << endl;
			cout << subtotal._height << endl;
		}
		total.append(subtotal, 'x');
	}
	uint16* buf = total.data();
	for (int32 j = 0; j < dataSize; ++j) {
		data[j] = buf[j];
		sum += (float64) data[j];
	}

	owner()->data.sendEvent();
	owner()->m_SynchroniseReadTerminated = true;
}

/*
 * read
 */
void RealPhotonicsDriver::read() {
}

/*
 * start
 */
void RealPhotonicsDriver::start() {

	// Set exposure time
	float64 exposure_time = owner()->time.setpoint() * 1000.;

    StartFeatureCommandExposure(&m_camera_handler, exposure_time);
     std::cout << "Exposure time successfully written to cameras. Please hit 4 to print current exposure setting." << std::endl;
     StartCaptureCommand(&m_camera_handler, 1);
     WaitUntilStreamingIsActive(&m_camera_handler);



//
//	const std::vector<CCameraControl*>* camera_vector = m_camera_handler.GetCameraControlVector();
//	for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin(); camera_iter != camera_vector->end();
//			camera_iter++) {
//		(*camera_iter)->StartFeatureCommandExposure(exposure_time);
//	}
//
//	// Start exposure
//	for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin(); camera_iter != camera_vector->end();
//			camera_iter++) {
//		(*camera_iter)->StartCaptureCommand(1);
//	}

//	bool all_cameras_active = true;
//	unsigned int repeats = 10;
//	unsigned int repeat_counter = 0;
//	std::stringstream cameras_streaming_status;
//	do {
//		cameras_streaming_status.str("");
//		all_cameras_active = true;
//		const std::vector<CCameraControl*>* camera_vector = m_camera_handler.GetCameraControlVector();
//		for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
//				camera_iter != camera_vector->end(); camera_iter++) {
//			if ((*camera_iter)->capture_active_) {
//				cameras_streaming_status << "Camera " << (*camera_iter)->camera_pointer_->GetModel() << "("
//						<< (*camera_iter)->camera_pointer_->GetDisplayName() << ")" << " successfully started" << std::endl;
//			} else {
//				cameras_streaming_status << "Camera " << (*camera_iter)->camera_pointer_->GetModel() << "("
//						<< (*camera_iter)->camera_pointer_->GetDisplayName() << ")" << " not started" << std::endl;
//				all_cameras_active = false;
//				break;
//			}
//		}
//		std::this_thread::sleep_for(std::chrono::milliseconds(100));
//	} while (!all_cameras_active && repeat_counter++ < repeats);
//	if (all_cameras_active) {
//		std::cout << "Capture command successfully started. Please hit 4 to print camera status." << std::endl;
//	} else {
//		std::cout << "Capture command not successfully. Not all camera are started correctly." << std::endl;
//	}
//	std::cout << cameras_streaming_status.str() << std::endl;
	m_TimeCounter.clear();
}

/*
 * resume
 */
void RealPhotonicsDriver::resume() {
}

/*
 * pause
 */
void RealPhotonicsDriver::pause() {

}

/*
 * stop
 */
void RealPhotonicsDriver::stop() {
}

/*
 * readStatus
 */
void RealPhotonicsDriver::readStatus() {
	int32 status = 0;
	int32 ret=0;
	float64 actualtime = m_TimeCounter.getTime();
	if (actualtime < owner()->time.setpoint()) {
		owner()->time = actualtime;
	} else {
		owner()->time = owner()->time.setpoint();
	}
	cout << "RealPhotonicsDriver::readStatus 1 " << endl;

	if (owner()->startActivated() == true) {
		const std::vector<CCameraControl*>* camera_vector = m_camera_handler.GetCameraControlVector();
		ret=0;
		for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
				camera_iter != camera_vector->end(); camera_iter++) {
			// perform logging output
			PrintLoggingInformation(*camera_iter);
			// perform status output
			PrintCameraStatus(*camera_iter);

			ret=ret+m_camera_handler.CapturedImages();
			cout << "RealPhotonicsDriver::readStatus 2  ret= " <<ret<< endl;
		}
		cout << "m_camera_handler.CapturedImages() = " << m_camera_handler.CapturedImages()
						<< " , m_camera_handler.NumberOfCameras() = " << m_camera_handler.NumberOfCameras() << endl;
		if (ret == 2) {
			cout << "RealPhotonicsDriver::readStatus 3 " << endl;
			status |= END;
			owner()->startActivated = false;
			owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
		} else {
			status |= COUNTING;
			owner()->calculateProgression();
		}
	} else {
		if (owner()->m_SynchroniseReadActivated == true) {
			if (owner()->m_SynchroniseReadTerminated == true) {
				owner()->m_SynchroniseReadActivated = false;
				owner()->m_SynchroniseReadTerminated = false;
				owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
			}
		} else {
			owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
		}
	}
	owner()->computeGateStatus(status);
	owner()->computeAcquisitionStatus(0);
	owner()->computeDetectorStatus(0);
}

/*
 * readInfos
 */
void RealPhotonicsDriver::readInfos() {

}



// this function sets the new data flag for all connected cameras to true, if the buffer_ pointer is nullptr,
// that means if all application threads have returned their camera buffer
// and a new camera buffer is served by the buffer handler
bool  RealPhotonicsDriver::HasDoubleBufferHandlerNewData(CDoubleBufferHandler* buffer_handler) {
    std::lock_guard<std::mutex> lock(current_buffer[buffer_handler].lock_);
    if (current_buffer[buffer_handler].buffer_ == nullptr) {
        if (buffer_handler->HasNewData()) {
            std::map<std::thread::id, std::atomic<bool>>::iterator it = current_buffer[buffer_handler].new_data.begin();
            for( ; it != current_buffer[buffer_handler].new_data.end(); it++) {
                (*it).second = true;
            }
        }
    }
    return current_buffer[buffer_handler].new_data[std::this_thread::get_id()];
}

// the function PullBufferFromDoubleBufferHandler uses a mutex to synchronize
// multiple calls to this function and to function FreeBufferToDoubleBufferHandler
// the mutex is entered, when this function is called and when the lock is not
// entered from FreeBufferToDoubleBufferHandler, in this case the current thread
// stops execution on the mutex object until FreeBufferToDoubleBufferHandler was left
BGAPI2::Buffer*  RealPhotonicsDriver::PullBufferFromDoubleBufferHandler(CDoubleBufferHandler* buffer_handler) {
    std::lock_guard<std::mutex> lock(current_buffer[buffer_handler].lock_);
    if (current_buffer[buffer_handler].buffer_ == nullptr) {
        current_buffer[buffer_handler].buffer_ = buffer_handler->PullBuffer();
    }
    if (current_buffer[buffer_handler].buffer_ != nullptr) {
        current_buffer[buffer_handler].ref_counter_++;
    }
    current_buffer[buffer_handler].new_data[std::this_thread::get_id()] = false;
    return current_buffer[buffer_handler].buffer_;
}

// vise versa to function PullBufferFromDoubleBufferHandler
void  RealPhotonicsDriver::FreeBufferToDoubleBufferHandler(CDoubleBufferHandler* buffer_handler, BGAPI2::Buffer* buffer) {
    std::lock_guard<std::mutex> lock(current_buffer[buffer_handler].lock_);
    if (buffer == current_buffer[buffer_handler].buffer_) {
        if (current_buffer[buffer_handler].ref_counter_ > 0) {
            current_buffer[buffer_handler].ref_counter_--;
        }
        if (current_buffer[buffer_handler].ref_counter_ == 0) {
            buffer_handler->FreeBuffer(buffer);
            current_buffer[buffer_handler].buffer_ = nullptr;
        }
    }
}

// -------------
// thread routines of application threads
// -------------

// the StatusThreadRoutine is left and the status thread finished, if the finish flag was set
// the StatusThreadRoutine starts working if the status command was activated from within the main loop
// the StatusThreadRoutine works with polling to detect an activated status command
void  RealPhotonicsDriver::StatusThreadRoutine(CCameraHandler* camera_handler) {
    // abort condition of the thread routine
    while (!finish_console_threads) {
        // check if the user activated the status command
        if (command_status) {
            const std::vector<CCameraControl*>* camera_vector = camera_handler->GetCameraControlVector();
            for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
                camera_iter != camera_vector->end();
                camera_iter++) {
                // perform logging output
                PrintLoggingInformation(*camera_iter);
                // perform status output
                PrintCameraStatus(*camera_iter);
            }
            if (command_calculation) {
                for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
                    camera_iter != camera_vector->end();
                    camera_iter++) {
                    PrintCalculationStatus((*camera_iter));
                }
            }

            // status command is now finished
            command_status = false;
        }
        else {
            // just wait until status command become active
            std::this_thread::sleep_for(std::chrono::milliseconds(10));
        }
    }
}

// the CalculationThreadRoutine works with polling to detect an activated calculation command
// the CalculationThreadRoutine is left and the calculation thread is finished, if the finish flag was set
// the CalculationThreadRoutine starts working if the calculation command was activated from within the main loop
void  RealPhotonicsDriver::CalculationThreadRoutine(CCameraHandler* camera_handler) {
    // the main loop of the thread, leaving this loop means finishing the thread
    // wait with the join function in the main thread until this loop was left
    while (!finish_console_threads) {
        // check if the user activated the status command
        if (command_calculation) {
        	// perform calculation
            const std::vector<CCameraControl*>* camera_vector = camera_handler->GetCameraControlVector();
            // iterate over all cameras and print the status messages to console
            for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
                camera_iter != camera_vector->end();
                camera_iter++) {
                DoCalculation(*camera_iter);
            }
            // wait a long time before proceeding, to show that more buffers are captured
            // then the calculation thread can process
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
        }
        else {
            // just wait until status command become active
            std::this_thread::sleep_for(std::chrono::milliseconds(10));
        }
    }
}

void  RealPhotonicsDriver::InitCurrentBufferStruct(CCameraHandler *camera_handler, std::list<std::thread::id> thread_id_list) {
    const std::vector<CCameraControl*>* camera_vector = camera_handler->GetCameraControlVector();
    for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
        camera_iter != camera_vector->end();
        camera_iter++) {
        for (std::list<std::thread::id>::iterator iter = thread_id_list.begin(); iter != thread_id_list.end(); iter++) {
            current_buffer[&(*camera_iter)->buffer_handler_].new_data[*iter] = true;
        }
    }
}

void  RealPhotonicsDriver::WaitUntilStreamingIsActive(CCameraHandler * camera_handler) {
    bool all_cameras_active = true;
    unsigned int repeats = 10;
    unsigned int repeat_counter = 0;
    std::stringstream cameras_streaming_status;
    do {
        cameras_streaming_status.str("");
        all_cameras_active = true;
        const std::vector<CCameraControl*>* camera_vector = camera_handler->GetCameraControlVector();
        for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
            camera_iter != camera_vector->end();
            camera_iter++) {
            if ((*camera_iter)->capture_active_) {
                cameras_streaming_status << "Camera " << (*camera_iter)->camera_pointer_->GetModel()
                    << "(" << (*camera_iter)->camera_pointer_->GetDisplayName() << ")"
                    << " successfully started" << std::endl;
            }
            else {
                cameras_streaming_status << "Camera " << (*camera_iter)->camera_pointer_->GetModel()
                    << "(" << (*camera_iter)->camera_pointer_->GetDisplayName() << ")"
                    << " not started" << std::endl;
                all_cameras_active = false;
                break;
            }
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(100));
    } while (!all_cameras_active && repeat_counter++ < repeats);
    if (all_cameras_active) {
        std::cout << "Capture command successfully started. Please hit 4 to print camera status." << std::endl;
    }
    else {
        std::cout << "Capture command not successfully. Not all camera are started correctly." << std::endl;
    }
    std::cout << cameras_streaming_status.str() << std::endl;

}

// this function waits until the expected number of images are captured or aborts
// if the time out time is reached
int RealPhotonicsDriver:: WaitOfImages(CCameraHandler * camera_handler, unsigned int number_of_images) {
    int return_code = 0;
    unsigned int timeout_count = 0;
    unsigned int last_numofimages = 0;
    unsigned int captured_numofimages = 0;
    do {
        captured_numofimages = camera_handler->CapturedImages();
        if (captured_numofimages != last_numofimages) {
            last_numofimages = captured_numofimages;
            timeout_count = 0;
            return_code = 0;
        }
        else {
            timeout_count++;
            return_code = 1;
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
        }
    } while (captured_numofimages < number_of_images && timeout_count < 100);
    return return_code;
}

void  RealPhotonicsDriver::StartFeatureCommandExposure(CCameraHandler * camera_handler, double exposure_time) {
    const std::vector<CCameraControl*>* camera_vector = camera_handler->GetCameraControlVector();
    for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
        camera_iter != camera_vector->end();
        camera_iter++) {
        (*camera_iter)->StartFeatureCommandExposure(exposure_time);
    }
}

void  RealPhotonicsDriver::StartCaptureCommand(CCameraHandler * camera_handler, unsigned int images) {
    const std::vector<CCameraControl*>* camera_vector = camera_handler->GetCameraControlVector();
    for (std::vector<CCameraControl*>::const_iterator camera_iter = camera_vector->begin();
        camera_iter != camera_vector->end();
        camera_iter++) {
        (*camera_iter)->StartCaptureCommand(images);
    }
}

void  RealPhotonicsDriver::PrintLoggingInformation(CCameraControl * camera_control) {
    while (!camera_control->LoggingListEmpty()) {
        std::cout << camera_control->LoggingString() << std::endl;
    };
}

// the PrintCameraStatus receive one camera buffer with buffer information from the passed camera
// and print this information to the console output
// the PrintCameraStatus use the synchronisation functions for a synchronized buffer access
// the PrintCameraStatus works only once on the same camera buffer on a camera (see HasDoubleBufferHandlerNewData)
void  RealPhotonicsDriver::PrintCameraStatus(CCameraControl * camera_control) {
    std::cout << "status of camera " << camera_control->camera_pointer_->GetModel() << std::endl;
    // check for a new camera buffer
    if (HasDoubleBufferHandlerNewData(&camera_control->buffer_handler_)) {
        // receive a camera buffer
        BGAPI2::Buffer * buffer = PullBufferFromDoubleBufferHandler(&camera_control->buffer_handler_);
        if (buffer) {
            BufferInformation * buffer_info = reinterpret_cast<BufferInformation*>(buffer->GetUserObj());
            if (!buffer->GetIsIncomplete()) {
                    if (camera_control->capture_active_) {
                        std::cout << "capturing is active" << std::endl;
                    }
                    else {
                        std::cout << "capturing is inactive" << std::endl;
                    }
                    std::cout << "number of captured images: " << camera_control->number_of_captured_images_ << std::endl;
                    std::cout << "number of incomplete frames: " << buffer_info->number_of_incomplete_images << std::endl;
                    std::cout << "resend requests: " << buffer_info->resend_requests
                        << " single: " << buffer_info->resend_requests_single
                        << " range: " << buffer_info->resend_requests_range << std::endl;
                    if (buffer_info->frameid > 0) {
                        std::cout << "last processed image id: " << buffer_info->frameid << std::endl;
                    }
            }
            else {
                std::cout << "buffer status skipped, current buffer was not completely received." << std::endl
                    << "captured images: " << camera_control->number_of_captured_images_ << std::endl;
            }
            // return the camera buffer
            FreeBufferToDoubleBufferHandler(&camera_control->buffer_handler_, buffer);
        }
        else {
            std::cout << "no new buffer captured: " << camera_control->number_of_captured_images_ << std::endl;
        }
    } else {
        std::cout << "no new buffer available" << std::endl
            << "captured images: " << camera_control->number_of_captured_images_ << std::endl;
    }
    std::stringstream outputstream;
    {
        std::lock_guard<std::mutex> lock(camera_control->feature_command_message_list_lock_);
        for (std::list<std::string>::iterator msg_iter = camera_control->feature_command_message_list_.begin();
            msg_iter != camera_control->feature_command_message_list_.end();
            msg_iter++) {
            outputstream << (*msg_iter) << std::endl;
        }
    }
    std::cout << outputstream.str() << std::endl;
    std::cout << std::endl << std::endl;
}

// print calculation results
void  RealPhotonicsDriver::PrintCalculationStatus(CCameraControl * camera_control) {
    std::lock_guard<std::mutex> lock(calculation_result_lock);
    BGAPI2::String camera_name = camera_control->camera_pointer_->GetModel();
    std::stringstream result_stream;
    result_stream << "Calculation result of" << std::endl <<
        "camera " << camera_name << std::endl <<
        "the next calculation is still running, no result available yet" << std::endl;

    if (calculation_result[&camera_control->buffer_handler_].size() == 0) {
        calculation_result[&camera_control->buffer_handler_] = result_stream.str();
    }
    std::cout << calculation_result[&camera_control->buffer_handler_] << std::endl;
    calculation_result[&camera_control->buffer_handler_] = result_stream.str();
}

// the DoCalculation receive one camera buffer and perform the calculation on the image data
// the DoCalculation use the synchronisation functions for a synchronized buffer access
// the DoCalculation works only once on a camera buffer (see HasDoubleBufferHandlerNewData)
// the DoCalculation writes the calculation result in a string
void  RealPhotonicsDriver::DoCalculation(CCameraControl * camera_control) {
    BGAPI2::String camera_name = camera_control->camera_pointer_->GetModel();
    std::stringstream result_stream;
   if (HasDoubleBufferHandlerNewData(&camera_control->buffer_handler_)) {
	    std::cout << "DoCalculation : " << camera_control->camera_pointer_->GetDisplayName() << std::endl;
        BGAPI2::Buffer * buffer = PullBufferFromDoubleBufferHandler(&camera_control->buffer_handler_);
        if (buffer) {
            BufferInformation * buffer_info = reinterpret_cast<BufferInformation*>(buffer->GetUserObj());
            if (!buffer->GetIsIncomplete()) {
                unsigned char* image_buffer = static_cast<unsigned char*>(buffer->GetMemPtr());
                int width = static_cast<int>(buffer->GetWidth());
                int height = static_cast<int>(buffer->GetHeight());
                std::cout << width << " , " << height << std::endl;
                int pixel_step_x = static_cast<int>(width / 200);
                int pixel_step_y = static_cast<int>((pixel_step_x * height) / width);
                int pixel_counter = 0;
                double pixel_sum = 0.0;
                for (int h = 0; h < height; h += pixel_step_y) {
                    for (int w = 0; w < width; w += pixel_step_x)
                    {
                        pixel_sum += image_buffer[h*width + w];
                        pixel_counter++;
                    }
                }
                double average = pixel_sum / pixel_counter;
                result_stream << "Calculation result of" << std::endl <<
                    "camera " << camera_name << std::endl <<
                    "last processed image id: " << buffer_info->frameid << std::endl <<
                    "mean gray value " << average << std::endl;
                std::lock_guard<std::mutex> lock(calculation_result_lock);
                calculation_result[&camera_control->buffer_handler_] = result_stream.str();
            }
            else {
                result_stream << "Calculation result of" << std::endl <<
                    "camera " << camera_name << std::endl <<
                    "calculation skipped, current buffer was not completely received " << std::endl;
                std::lock_guard<std::mutex> lock(calculation_result_lock);
                calculation_result[&camera_control->buffer_handler_] = result_stream.str();
            }

            FreeBufferToDoubleBufferHandler(&camera_control->buffer_handler_, buffer);
        }
        else {
            std::lock_guard<std::mutex> lock(calculation_result_lock);
            calculation_result[&camera_control->buffer_handler_] = "No image for calculation captured!";
        }
    }
    else {
        std::stringstream out;
        out << "Calculation result of" << std::endl <<
            "camera " << camera_name << std::endl <<
            "threadId " << std::this_thread::get_id() << std::endl <<
            "no new buffer available: " << camera_control->number_of_captured_images_ << std::endl;
        std::lock_guard<std::mutex> lock(calculation_result_lock);
        calculation_result[&camera_control->buffer_handler_] = result_stream.str();
    }

}







}
