/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef STEREOELECCONTROLLER_H
#define STEREOELECCONTROLLER_H

#include <Controller.h>
#include "StereoElecDriver.h"

namespace stereo_elec {


/*!
 * \class StereoElecController
 */

class StereoElecController: public ExperimentController{

public:

	//! Type of controller
	static const std::string TYPE;

	/*!
	 * \brief Constructor
	 *
	 * It is the default constructor called by the creator class ExperimentControllerCreator.
	 *
	 * \param[in] name the name of the controller
	 */
	StereoElecController(const std::string& name);

	/*!
	 * \brief Copy constructor
	 *
	 * \param[in] controller the controller to copy
	 */
	StereoElecController(const StereoElecController& controller);

	/*!
	 * \brief Destructor
	 */
	virtual ~StereoElecController();

	/*!
	 * \brief Method called before changing the property value
	 *
	 * This method is called after setting configuration during the creation of controller.
	 */
	virtual void postConfiguration();

	/*!
	 * Properties
	 */




protected:
		Property<int32> nbfemax;
		Property<int32> nbledmax;
		// FE, size NBFEMAX
		DynamicProperty<int32> slotindex;
		DynamicProperty<float64> fe_pos5v; /**< Front-end positive power supply. */
		DynamicProperty<float64> fe_neg5v; /**< Front-end negative power supply. */
		DynamicProperty<float64> fe_adc3v3a; /**< ADC power supply. */
		DynamicProperty<float64> fe_adc1v8a;/**< ADC power supply. */
		DynamicProperty<float64> fe_tadc03;/**< Temperature probe close to ADC0-3.*/
		DynamicProperty<float64> fe_tadc47;/**< Temperature probe close to ADC4-7.*/
		// LED, size NBLEDMAX
		DynamicProperty<float64> led_dac3v3; /**< DAC setting power supply */
		DynamicProperty<float64> led_amp24v; /**< Amplifier supply. */

		// Properties for rates
		DynamicProperty<float64> rate0;
		DynamicProperty<float64> rate1;
		DynamicProperty<float64> rate2;
		DynamicProperty<float64> rate3;
		DynamicProperty<float64> rate4;
		DynamicProperty<float64> rate5;
		DynamicProperty<float64> rate6;
		DynamicProperty<float64> rate7;

		Property<int32> t1accepted;
		Property<int32> t1rejected;
		Property<int32> fiforejected;
		Property<int32> t2accepted;
		Property<float64> t1rejectedfraction;
		Property<std::string> runname;

		DriverPtr<StereoElecDriver> m_Driver; //! The driver linked to the controller




private:
	void refreshNbFEMaxProperty(int32 value) throw (CannotSetValue);
	void refreshNbLEDMaxProperty(int32 value) throw (CannotSetValue);
	virtual void updateFE_Pos5V(int32 index);
	virtual void updateFE_Neg5V(int32 index);
	virtual void updateFE_ADC1V8A(int32 index);
	virtual void updateFE_ADC3V3A(int32 index);
	virtual void updateFE_TADC03(int32 index);
	virtual void updateFE_TADC47(int32 index);
	virtual void updateLED_DAC3V3(int32 index);
	virtual void updateLED_Amp24V(int32 index);

	virtual void updateRate0(int32 index);
	virtual void updateRate1(int32 index);
	virtual void updateRate2(int32 index);
	virtual void updateRate3(int32 index);
	virtual void updateRate4(int32 index);
	virtual void updateRate5(int32 index);
	virtual void updateRate6(int32 index);
	virtual void updateRate7(int32 index);

	virtual void updateT1Accepted();
	virtual void updateT1Rejected();
	virtual void updateFIFORejected();
	virtual void updateT2Accepted();
	virtual void updateRunName();
/*!
	 * \brief update the progression property from driver one
	 */
	void updateProgression();
};
}
#endif // STEREOELECCONTROLLER_H
